# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Tuple

from mmocr.registry import DATA_GATHERERS
from .base import BaseGatherer


@DATA_GATHERERS.register_module()
class MonoGatherer(BaseGatherer):
    """Gather the dataset file. Specifically for the case that only one
    annotation file is needed. For example,

            img_001.jpg \
            img_002.jpg ---> train.json
            img_003.jpg /

    Args:
        ann_name (str): The name of the annotation file.
    """

    def __init__(self, ann_name: str, **kwargs) -> None:
        super().__init__(**kwargs)

        self.ann_name = ann_name

    def __call__(self) -> Tuple[str, str]:
        """
        Returns:
            tuple(str, str): The directory of the image and the path of
            annotation file.
        """

        return (self.img_dir, osp.join(self.ann_dir, self.ann_name))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict

from mmcv.transforms.base import BaseTransform
from mmdet.structures.mask import PolygonMasks, bitmap_to_polygon

from mmocr.registry import TRANSFORMS


@TRANSFORMS.register_module()
class MMDet2MMOCR(BaseTransform):
    """Convert transforms's data format from MMDet to MMOCR.

    Required Keys:

    - gt_masks (PolygonMasks | BitmapMasks) (optional)
    - gt_ignore_flags (np.bool) (optional)

    Added Keys:

    - gt_polygons (list[np.ndarray])
    - gt_ignored (np.ndarray)
    """

    def transform(self, results: Dict) -> Dict:
        """Convert MMDet's data format to MMOCR's data format.

        Args:
            results (Dict): Result dict containing the data to transform.

        Returns:
            (Dict): The transformed data.
        """
        # gt_masks -> gt_polygons
        if 'gt_masks' in results.keys():
            gt_polygons = []
            gt_masks = results.pop('gt_masks')
            if len(gt_masks) > 0:
                # PolygonMasks
                if isinstance(gt_masks[0], PolygonMasks):
                    gt_polygons = [mask[0] for mask in gt_masks.masks]
                # BitmapMasks
                else:
                    polygons = []
                    for mask in gt_masks.masks:
                        contours, _ = bitmap_to_polygon(mask)
                        polygons += [
                            contour.reshape(-1) for contour in contours
                        ]
                    # filter invalid polygons
                    gt_polygons = []
                    for polygon in polygons:
                        if len(polygon) < 6:
                            continue
                        gt_polygons.append(polygon)

            results['gt_polygons'] = gt_polygons
        # gt_ignore_flags -> gt_ignored
        if 'gt_ignore_flags' in results.keys():
            gt_ignored = results.pop('gt_ignore_flags')
            results['gt_ignored'] = gt_ignored

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        return repr_str


@TRANSFORMS.register_module()
class MMOCR2MMDet(BaseTransform):
    """Convert transforms's data format from MMOCR to MMDet.

    Required Keys:

    - img_shape
    - gt_polygons (List[ndarray]) (optional)
    - gt_ignored (np.bool) (optional)

    Added Keys:

    - gt_masks (PolygonMasks | BitmapMasks) (optional)
    - gt_ignore_flags (np.bool) (optional)

    Args:
        poly2mask (bool): Whether to convert mask to bitmap. Default: True.
    """

    def __init__(self, poly2mask: bool = False) -> None:
        self.poly2mask = poly2mask

    def transform(self, results: Dict) -> Dict:
        """Convert MMOCR's data format to MMDet's data format.

        Args:
            results (Dict): Result dict containing the data to transform.

        Returns:
            (Dict): The transformed data.
        """
        # gt_polygons -> gt_masks
        if 'gt_polygons' in results.keys():
            gt_polygons = results.pop('gt_polygons')
            gt_polygons = [[gt_polygon] for gt_polygon in gt_polygons]
            gt_masks = PolygonMasks(gt_polygons, *results['img_shape'])

            if self.poly2mask:
                gt_masks = gt_masks.to_bitmap()

            results['gt_masks'] = gt_masks
        # gt_ignore_flags -> gt_ignored
        if 'gt_ignored' in results.keys():
            gt_ignored = results.pop('gt_ignored')
            results['gt_ignore_flags'] = gt_ignored

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(poly2mask = {self.poly2mask})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn

from mmocr.models.common.dictionary import Dictionary
from mmocr.registry import MODELS
from mmocr.structures import TextRecogDataSample
from .base import BaseDecoder


@MODELS.register_module()
class ASTERDecoder(BaseDecoder):
    """Implement attention decoder.

    Args:
        in_channels (int): Number of input channels.
        emb_dims (int): Dims of char embedding. Defaults to 512.
        attn_dims (int): Dims of attention. Both hidden states and features
            will be projected to this dims. Defaults to 512.
        hidden_size (int): Dims of hidden state for GRU. Defaults to 512.
        dictionary (dict or :obj:`Dictionary`): The config for `Dictionary` or
            the instance of `Dictionary`. Defaults to None.
        max_seq_len (int): Maximum output sequence length :math:`T`. Defaults
            to 25.
        module_loss (dict, optional): Config to build loss. Defaults to None.
        postprocessor (dict, optional): Config to build postprocessor.
            Defaults to None.
        init_cfg (dict or list[dict], optional): Initialization configs.
            Defaults to None.
    """

    def __init__(self,
                 in_channels: int,
                 emb_dims: int = 512,
                 attn_dims: int = 512,
                 hidden_size: int = 512,
                 dictionary: Union[Dictionary, Dict] = None,
                 max_seq_len: int = 25,
                 module_loss: Dict = None,
                 postprocessor: Dict = None,
                 init_cfg=dict(type='Xavier', layer='Conv2d')):
        super().__init__(
            init_cfg=init_cfg,
            dictionary=dictionary,
            module_loss=module_loss,
            postprocessor=postprocessor,
            max_seq_len=max_seq_len)

        self.start_idx = self.dictionary.start_idx
        self.num_classes = self.dictionary.num_classes
        self.in_channels = in_channels
        self.embedding_dim = emb_dims
        self.att_dims = attn_dims
        self.hidden_size = hidden_size

        # Projection layers
        self.proj_feat = nn.Linear(in_channels, attn_dims)
        self.proj_hidden = nn.Linear(hidden_size, attn_dims)
        self.proj_sum = nn.Linear(attn_dims, 1)

        # Decoder input embedding
        self.embedding = nn.Embedding(self.num_classes, self.att_dims)

        # GRU
        self.gru = nn.GRU(
            input_size=self.in_channels + self.embedding_dim,
            hidden_size=self.hidden_size,
            batch_first=True)

        # Prediction layer
        self.fc = nn.Linear(hidden_size, self.dictionary.num_classes)
        self.softmax = nn.Softmax(dim=-1)

    def _attention(self, feat: torch.Tensor, prev_hidden: torch.Tensor,
                   prev_char: torch.Tensor
                   ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Implement the attention mechanism.

        Args:
            feat (Tensor): Feature map from encoder of shape :math:`(N, T, C)`.
            prev_hidden (Tensor): Previous hidden state from GRU of shape
                :math:`(1, N, self.hidden_size)`.
            prev_char (Tensor): Previous predicted character of shape
                :math:`(N, )`.

        Returns:
            tuple(Tensor, Tensor):
            - output (Tensor): Predicted character of current time step of
              shape :math:`(N, 1)`.
            - state (Tensor): Hidden state from GRU of current time step of
              shape :math:`(N, self.hidden_size)`.
        """
        # Calculate the attention weights
        B, T, _ = feat.size()
        feat_proj = self.proj_feat(feat)  # [N, T, attn_dims]
        hidden_proj = self.proj_hidden(prev_hidden)  # [1, N, attn_dims]
        hidden_proj = hidden_proj.squeeze(0).unsqueeze(1)  # [N, 1, attn_dims]
        hidden_proj = hidden_proj.expand(B, T,
                                         self.att_dims)  # [N, T, attn_dims]

        sum_tanh = torch.tanh(feat_proj + hidden_proj)  # [N, T, attn_dims]
        sum_proj = self.proj_sum(sum_tanh).squeeze(-1)  # [N, T]
        attn_weights = torch.softmax(sum_proj, dim=1)  # [N, T]

        # GRU forward
        context = torch.bmm(attn_weights.unsqueeze(1), feat).squeeze(1)
        char_embed = self.embedding(prev_char.long())  # [N, emb_dims]
        output, state = self.gru(
            torch.cat([char_embed, context], 1).unsqueeze(1), prev_hidden)
        output = output.squeeze(1)
        output = self.fc(output)
        return output, state

    def forward_train(
        self,
        feat: torch.Tensor = None,
        out_enc: Optional[torch.Tensor] = None,
        data_samples: Optional[Sequence[TextRecogDataSample]] = None
    ) -> torch.Tensor:
        """
        Args:
            feat (Tensor): Feature from backbone. Unused in this decoder.
            out_enc (torch.Tensor, optional): Encoder output. Defaults to None.
            data_samples (list[TextRecogDataSample], optional): Batch of
                TextRecogDataSample, containing gt_text information. Defaults
                to None.

        Returns:
            Tensor: The raw logit tensor. Shape :math:`(N, T, C)` where
            :math:`C` is ``num_classes``.
        """
        B = out_enc.shape[0]
        state = torch.zeros(1, B, self.hidden_size).to(out_enc.device)
        padded_targets = [
            data_sample.gt_text.padded_indexes for data_sample in data_samples
        ]
        padded_targets = torch.stack(padded_targets, dim=0).to(out_enc.device)
        outputs = []
        for i in range(self.max_seq_len):
            prev_char = padded_targets[:, i].to(out_enc.device)
            output, state = self._attention(out_enc, state, prev_char)
            outputs.append(output)
        outputs = torch.cat([_.unsqueeze(1) for _ in outputs], 1)
        return outputs

    def forward_test(
        self,
        feat: Optional[torch.Tensor] = None,
        out_enc: Optional[torch.Tensor] = None,
        data_samples: Optional[Sequence[TextRecogDataSample]] = None
    ) -> torch.Tensor:
        """
        Args:
            feat (Tensor): Feature from backbone. Unused in this decoder.
            out_enc (torch.Tensor, optional): Encoder output. Defaults to None.
            data_samples (list[TextRecogDataSample], optional): Batch of
                TextRecogDataSample, containing gt_text information. Defaults
                to None. Unused in this decoder.

        Returns:
            Tensor: The raw logit tensor. Shape :math:`(N, T, C)` where
            :math:`C` is ``num_classes``.
        """
        B = out_enc.shape[0]
        predicted = []
        state = torch.zeros(1, B, self.hidden_size).to(out_enc.device)
        outputs = []
        for i in range(self.max_seq_len):
            if i == 0:
                prev_char = torch.zeros(B).fill_(self.start_idx).to(
                    out_enc.device)
            else:
                prev_char = predicted

            output, state = self._attention(out_enc, state, prev_char)
            outputs.append(output)
            _, predicted = output.max(-1)
        outputs = torch.cat([_.unsqueeze(1) for _ in outputs], 1)
        return self.softmax(outputs)



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import List

from mmocr.registry import DATA_DUMPERS
from mmocr.utils import list_to_file
from .base import BaseDumper


@DATA_DUMPERS.register_module()
class WildreceiptOpensetDumper(BaseDumper):

    def dump(self, data: List):
        """Dump data to txt file.

        Args:
            data (List): Data to be dumped.
        """

        filename = f'openset_{self.split}.txt'
        dst_file = osp.join(self.data_root, filename)
        list_to_file(dst_file, data)



_base_ = [
    'mask-rcnn_resnet50_fpn_160e_ctw1500.py',
]

load_from = None

_base_.model.cfg.backbone = dict(
    _scope_='mmocr',
    type='CLIPResNet',
    init_cfg=dict(
        type='Pretrained',
        checkpoint='https://download.openmmlab.com/'
        'mmocr/backbone/resnet50-oclip-7ba0c533.pth'))

_base_.optim_wrapper.optimizer.lr = 0.02



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp

from mmocr.utils import dump_ocr_data


def convert_annotations(root_path, split):
    """Convert original annotations to mmocr format.

    The annotation format of this dataset is as the following:
        word_1.png, "flying"
        word_2.png, "today"
        word_3.png, "means"
    See the format of converted annotation in mmocr.utils.dump_ocr_data.

    Args:
        root_path (str): The root path of the dataset
        split (str): The split of dataset. Namely: Train or Test
    """
    assert isinstance(root_path, str)
    assert isinstance(split, str)

    img_info = []
    with open(
            osp.join(root_path, 'annotations',
                     f'Challenge1_{split}_Task3_GT.txt'),
            encoding='"utf-8-sig') as f:
        annos = f.readlines()
    for anno in annos:
        # text may contain comma ','
        dst_img_name, word = anno.split(', "')
        word = word.replace('"\n', '')

        img_info.append({
            'file_name': dst_img_name,
            'anno_info': [{
                'text': word
            }]
        })

    return img_info


def parse_args():
    parser = argparse.ArgumentParser(
        description='Generate training and test set of IC11')
    parser.add_argument('root_path', help='Root dir path of IC11')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    root_path = args.root_path

    for split in ['Train', 'Test']:
        img_info = convert_annotations(root_path, split)
        dump_ocr_data(img_info,
                      osp.join(root_path, f'{split.lower()}_label.json'),
                      'textrecog')
        print(f'{split} split converted.')


if __name__ == '__main__':
    main()



icdar2015_textrecog_data_root = 'data/icdar2015'

icdar2015_textrecog_train = dict(
    type='OCRDataset',
    data_root=icdar2015_textrecog_data_root,
    ann_file='textrecog_train.json',
    pipeline=None)

icdar2015_textrecog_test = dict(
    type='OCRDataset',
    data_root=icdar2015_textrecog_data_root,
    ann_file='textrecog_test.json',
    test_mode=True,
    pipeline=None)

icdar2015_1811_textrecog_test = dict(
    type='OCRDataset',
    data_root=icdar2015_textrecog_data_root,
    ann_file='textrecog_test_1811.json',
    test_mode=True,
    pipeline=None)



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import List, Optional, Tuple, Union


class BaseGatherer:
    """Base class for gatherer.

    Note: Gatherer assumes that all the annotation file is in the same
    directory and all the image files are in the same directory.

    Args:
        img_dir(str): The directory of the images. It is usually set
            automatically to f'text{task}_imgs/split' and users do not need to
            set it manually in config file in most cases. When the image files
            is not in 'text{task}_imgs/split' directory, users should set it.
            Defaults to ''.
        ann_dir (str): The directory of the annotation files. It is usually set
            automatically to 'annotations' and users do not need to set it
            manually in config file in most cases. When the annotation files
            is not in 'annotations' directory, users should set it. Defaults to
            'annotations'.
        split (str, optional): List of splits to gather. It' s the partition of
            the datasets. Options are 'train', 'val' or 'test'. It is usually
            set automatically and users do not need to set it manually in
            config file in most cases. Defaults to None.
        data_root (str, optional): The root directory of the image and
            annotation. It is usually set automatically and users do not need
            to set it manually in config file in most cases. Defaults to None.
    """

    def __init__(self,
                 img_dir: str = '',
                 ann_dir: str = 'annotations',
                 split: Optional[str] = None,
                 data_root: Optional[str] = None) -> None:
        self.split = split
        self.data_root = data_root
        self.ann_dir = osp.join(data_root, ann_dir)
        self.img_dir = osp.join(data_root, img_dir)

    def __call__(self) -> Union[Tuple[List[str], List[str]], Tuple[str, str]]:
        """The return value of the gatherer is a tuple of two lists or strings.

        The first element is the list of image paths or the directory of the
        images. The second element is the list of annotation paths or the path
        of the annotation file which contains all the annotations.
        """
        raise NotImplementedError



# Copyright (c) OpenMMLab. All rights reserved.
from .db_module_loss import DBModuleLoss
from .drrg_module_loss import DRRGModuleLoss
from .fce_module_loss import FCEModuleLoss
from .pan_module_loss import PANModuleLoss
from .pse_module_loss import PSEModuleLoss
from .seg_based_module_loss import SegBasedModuleLoss
from .textsnake_module_loss import TextSnakeModuleLoss

__all__ = [
    'PANModuleLoss', 'PSEModuleLoss', 'DBModuleLoss', 'TextSnakeModuleLoss',
    'FCEModuleLoss', 'DRRGModuleLoss', 'SegBasedModuleLoss'
]



# Copyright (c) OpenMMLab. All rights reserved.

from typing import Dict

import torch

from mmocr.registry import MODELS
from mmocr.utils.typing_utils import (ConfigType, InitConfigType,
                                      OptConfigType, OptRecSampleList,
                                      RecForwardResults, RecSampleList)
from .base import BaseRecognizer


@MODELS.register_module()
class EncoderDecoderRecognizer(BaseRecognizer):
    """Base class for encode-decode recognizer.

    Args:
        preprocessor (dict, optional): Config dict for preprocessor. Defaults
            to None.
        backbone (dict, optional): Backbone config. Defaults to None.
        encoder (dict, optional): Encoder config. If None, the output from
            backbone will be directly fed into ``decoder``. Defaults to None.
        decoder (dict, optional): Decoder config. Defaults to None.
        data_preprocessor (dict, optional): Model preprocessing config
            for processing the input image data. Keys allowed are
            ``to_rgb``(bool), ``pad_size_divisor``(int), ``pad_value``(int or
            float), ``mean``(int or float) and ``std``(int or float).
            Preprcessing order: 1. to rgb; 2. normalization 3. pad.
            Defaults to None.
        init_cfg (dict or list[dict], optional): Initialization configs.
            Defaults to None.
    """

    def __init__(self,
                 preprocessor: OptConfigType = None,
                 backbone: OptConfigType = None,
                 encoder: OptConfigType = None,
                 decoder: OptConfigType = None,
                 data_preprocessor: ConfigType = None,
                 init_cfg: InitConfigType = None) -> None:

        super().__init__(
            init_cfg=init_cfg, data_preprocessor=data_preprocessor)

        # Preprocessor module, e.g., TPS
        if preprocessor is not None:
            self.preprocessor = MODELS.build(preprocessor)

        # Backbone
        if backbone is not None:
            self.backbone = MODELS.build(backbone)

        # Encoder module
        if encoder is not None:
            self.encoder = MODELS.build(encoder)

        # Decoder module
        assert decoder is not None
        self.decoder = MODELS.build(decoder)

    def extract_feat(self, inputs: torch.Tensor) -> torch.Tensor:
        """Directly extract features from the backbone."""
        if self.with_preprocessor:
            inputs = self.preprocessor(inputs)
        if self.with_backbone:
            inputs = self.backbone(inputs)
        return inputs

    def loss(self, inputs: torch.Tensor, data_samples: RecSampleList,
             **kwargs) -> Dict:
        """Calculate losses from a batch of inputs and data samples.
        Args:
            inputs (tensor): Input images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.
            data_samples (list[TextRecogDataSample]): A list of N
                datasamples, containing meta information and gold
                annotations for each of the images.

        Returns:
            dict[str, tensor]: A dictionary of loss components.
        """
        feat = self.extract_feat(inputs)
        out_enc = None
        if self.with_encoder:
            out_enc = self.encoder(feat, data_samples)
        return self.decoder.loss(feat, out_enc, data_samples)

    def predict(self, inputs: torch.Tensor, data_samples: RecSampleList,
                **kwargs) -> RecSampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (torch.Tensor): Image input tensor.
            data_samples (list[TextRecogDataSample]): A list of N datasamples,
                containing meta information and gold annotations for each of
                the images.

        Returns:
            list[TextRecogDataSample]:  A list of N datasamples of prediction
            results. Results are stored in ``pred_text``.
        """
        feat = self.extract_feat(inputs)
        out_enc = None
        if self.with_encoder:
            out_enc = self.encoder(feat, data_samples)
        return self.decoder.predict(feat, out_enc, data_samples)

    def _forward(self,
                 inputs: torch.Tensor,
                 data_samples: OptRecSampleList = None,
                 **kwargs) -> RecForwardResults:
        """Network forward process. Usually includes backbone, encoder and
        decoder forward without any post-processing.

         Args:
            inputs (Tensor): Inputs with shape (N, C, H, W).
            data_samples (list[TextRecogDataSample]): A list of N
                datasamples, containing meta information and gold
                annotations for each of the images.

        Returns:
            Tensor: A tuple of features from ``decoder`` forward.
        """
        feat = self.extract_feat(inputs)
        out_enc = None
        if self.with_encoder:
            out_enc = self.encoder(feat, data_samples)
        return self.decoder(feat, out_enc, data_samples)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence, Union

import torch
import torch.nn as nn
from mmengine.model import BaseModule

from mmocr.registry import MODELS


@MODELS.register_module()
class NRTRModalityTransform(BaseModule):
    """Modality transform in NRTR.

    Args:
        in_channels (int): Input channel of image. Defaults to 3.
        init_cfg (dict or list[dict], optional): Initialization configs.
    """

    def __init__(
        self,
        in_channels: int = 3,
        init_cfg: Optional[Union[Dict, Sequence[Dict]]] = [
            dict(type='Kaiming', layer='Conv2d'),
            dict(type='Uniform', layer='BatchNorm2d')
        ]
    ) -> None:
        super().__init__(init_cfg=init_cfg)

        self.conv_1 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=32,
            kernel_size=3,
            stride=2,
            padding=1)
        self.relu_1 = nn.ReLU(True)
        self.bn_1 = nn.BatchNorm2d(32)

        self.conv_2 = nn.Conv2d(
            in_channels=32,
            out_channels=64,
            kernel_size=3,
            stride=2,
            padding=1)
        self.relu_2 = nn.ReLU(True)
        self.bn_2 = nn.BatchNorm2d(64)

        self.linear = nn.Linear(512, 512)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Backbone forward.

        Args:
            x (torch.Tensor): Image tensor of shape :math:`(N, C, W, H)`. W, H
                is the width and height of image.
        Return:
            Tensor: Output tensor.
        """
        x = self.conv_1(x)
        x = self.relu_1(x)
        x = self.bn_1(x)

        x = self.conv_2(x)
        x = self.relu_2(x)
        x = self.bn_2(x)

        n, c, h, w = x.size()

        x = x.permute(0, 3, 2, 1).contiguous().view(n, w, h * c)

        x = self.linear(x)

        x = x.permute(0, 2, 1).contiguous().view(n, -1, 1, w)

        return x



# Copyright (c) OpenMMLab. All rights reserved.
from .conv_layer import BasicBlock, Bottleneck
from .dot_product_attention_layer import DotProductAttentionLayer
from .lstm_layer import BidirectionalLSTM
from .position_aware_layer import PositionAwareLayer
from .robust_scanner_fusion_layer import RobustScannerFusionLayer
from .satrn_layers import Adaptive2DPositionalEncoding, SATRNEncoderLayer

__all__ = [
    'BidirectionalLSTM', 'Adaptive2DPositionalEncoding', 'BasicBlock',
    'Bottleneck', 'RobustScannerFusionLayer', 'DotProductAttentionLayer',
    'PositionAwareLayer', 'SATRNEncoderLayer'
]
